package com.flow.service.impl;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flow.common.core.exception.BaseException;
import com.flow.common.core.model.PageResult;
import com.flow.common.core.util.StrUtil;
import com.flow.common.mybatis.service.impl.BaseServiceImpl;
import com.flow.dao.FlowInstanceDao;
import com.flow.entity.FlowDefine;
import com.flow.entity.FlowInstance;
import com.flow.entity.Notify;
import com.flow.entity.User;
import com.flow.entity.settings.PrintConfig;
import com.flow.entity.settings.Settings;
import com.flow.enums.NotifyEnum;
import com.flow.enums.ProcessStatus;
import com.flow.model.FlowInstanceQuery;
import com.flow.model.FormInfo;
import com.flow.model.spreadsheet.SpreadSheet;
import com.flow.service.*;
import org.flowable.engine.HistoryService;
import org.flowable.engine.RuntimeService;
import org.flowable.engine.TaskService;
import org.flowable.engine.history.HistoricProcessInstance;
import org.flowable.engine.runtime.ProcessInstance;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

@Service
public class FlowInstanceServiceImpl extends BaseServiceImpl<FlowInstanceDao, FlowInstance> implements FlowInstanceService {
    @Autowired
    private FlowInstanceDao flowInstanceDao;
    @Autowired
    private FlowDefineService flowDefineService;
    @Autowired
    private RuntimeService runtimeService;
    @Autowired
    private TaskService taskService;
    @Autowired
    private FlowTaskService flowTaskService;
    @Autowired
    private NotifyService notifyService;
    @Autowired
    private UserService userService;
    @Autowired
    private HistoryService historyService;
    @Autowired
    private ObjectMapper objectMapper;

    @Override
    public PageResult<FlowInstance> list(FlowInstanceQuery flowInstanceQuery) {
        Page<FlowInstance> page = new Page<>(
                flowInstanceQuery.getPage(),
                flowInstanceQuery.getLimit()
        );
        IPage<FlowInstance> iPage = flowInstanceDao.list(page, flowInstanceQuery);
        return new PageResult<>(iPage.getTotal(), iPage.getRecords());
    }

    @Override
    public FlowInstance getInstance(String instanceId) {
        return flowInstanceDao.getInstance(instanceId);
    }

    @Override
    public Map<String, Object> getVariables(String instanceId) {
        Map<String, Object> variables = new HashMap<>();
        ProcessInstance processInstance = runtimeService.createProcessInstanceQuery().processInstanceId(instanceId).singleResult();
        if (Objects.nonNull(processInstance)) {
            runtimeService.getVariableInstances(instanceId).values().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getName(), variableInstance.getValue())
            );
            return variables;
        }
        HistoricProcessInstance historicProcessInstance = historyService.createHistoricProcessInstanceQuery().processInstanceId(instanceId).singleResult();
        if (Objects.nonNull(historicProcessInstance)) {
            historyService.createHistoricVariableInstanceQuery().processInstanceId(instanceId).list().forEach(variableInstance ->
                    variables.putIfAbsent(variableInstance.getVariableName(), variableInstance.getValue())
            );
        }
        return variables;
    }

    @Override
    public FormInfo getForm(String instanceId) {
        FlowInstance instance = this.getInstance(instanceId);
        if (Objects.isNull(instance)) {
            throw new BaseException("流程实例不存在");
        }
        FlowDefine define = flowDefineService.getDefine(instance.getDefinitionId());
        if (Objects.isNull(define)) {
            throw new BaseException("流程不存在");
        }
        Map<String, Object> variables = this.getVariables(instanceId);
        FormInfo formInfo = new FormInfo(define, null, variables);
        formInfo.setInstanceId(instanceId);
        formInfo.setFormName(instance.getName());
        formInfo.setStartUserId(instance.getStartUserId());
        formInfo.setStartTime(instance.getStartTime());
        formInfo.setEndTime(instance.getEndTime());
        formInfo.setStatus(instance.getStatus());
        return formInfo;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void cancel(List<String> instanceIds) {
        for (String instanceId : instanceIds) {
            FlowInstance instance = this.getInstance(instanceId);
            if (instance.getStatus() != ProcessStatus.RUNNING) {
                throw new BaseException("流程已结束，不能撤回");
            }
            Integer cancelDays = instance.getCancelDays();
            if (cancelDays <= 0) {
                throw new BaseException("无法撤回流程");
            }
            LocalDateTime startTime = instance.getStartTime();
            startTime.plusDays(cancelDays);
            if (LocalDateTime.now().isAfter(startTime)) {
                throw new BaseException("流程撤回时间已过");
            }
            taskService.createTaskQuery().processInstanceId(instanceId).list().forEach(task -> {
                flowTaskService.updateStatus(task.getId(), ProcessStatus.CANCELED);
            });
            runtimeService.updateBusinessStatus(instanceId, ProcessStatus.CANCELED.toString());
            runtimeService.deleteProcessInstance(instanceId, "撤回");
            FlowInstance flowInstance = FlowInstance.buildStatus(instanceId, ProcessStatus.CANCELED);
            this.updateById(flowInstance);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void reject(String instanceId) {
        ProcessInstance processInstance = runtimeService.createProcessInstanceQuery()
                .processInstanceId(instanceId)
                .singleResult();
        if (Objects.isNull(processInstance)) {
            throw new BaseException("流程实例不存在");
        }
        runtimeService.updateBusinessStatus(instanceId, ProcessStatus.REFUSE.toString());
        runtimeService.deleteProcessInstance(instanceId, "拒绝");
        FlowInstance flowInstance = FlowInstance.buildStatus(instanceId, ProcessStatus.REFUSE);
        flowInstanceDao.updateById(flowInstance);
        // 消息通知
        Notify notify = new Notify();
        notify.setSubject("申请拒绝");
        notify.setContent(String.format("您申请的流程《%s》已拒绝", processInstance.getName()));
        notify.setSender("admin");
        notify.setType(NotifyEnum.SYSTEM);
        notify.setReceiver(processInstance.getStartUserId());
        notify.setReceivingTime(LocalDateTime.now());
        notify.setUrl("/flow/instance");
        notifyService.notify(notify);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void suspend(String instanceId) {
        ProcessInstance processInstance = runtimeService.createProcessInstanceQuery()
                .processInstanceId(instanceId)
                .singleResult();
        if (Objects.isNull(processInstance)) {
            throw new BaseException("流程实例不存在");
        }
        ProcessStatus status = processInstance.isSuspended() ? ProcessStatus.HANGUP : ProcessStatus.RUNNING;
        taskService.createTaskQuery().processInstanceId(instanceId).list().forEach(task -> {
            flowTaskService.updateStatus(task.getId(), status);
        });
        if (processInstance.isSuspended()) {
            runtimeService.activateProcessInstanceById(instanceId);
            runtimeService.updateBusinessStatus(instanceId, ProcessStatus.RUNNING.toString());
            FlowInstance flowInstance = FlowInstance.buildStatus(instanceId, ProcessStatus.RUNNING);
            this.updateById(flowInstance);
        } else {
            runtimeService.suspendProcessInstanceById(instanceId);
            runtimeService.updateBusinessStatus(instanceId, ProcessStatus.HANGUP.toString());
            FlowInstance flowInstance = FlowInstance.buildStatus(instanceId, ProcessStatus.HANGUP);
            this.updateById(flowInstance);
        }
    }

    @Override
    public SpreadSheet getPrintData(String instanceId) throws JsonProcessingException {
        FlowInstance instance = flowInstanceDao.getInstance(instanceId);
        String definitionId = instance.getDefinitionId();
        FlowDefine define = flowDefineService.getDefine(definitionId);
        if (Objects.isNull(define)) {
            throw new BaseException("流程不存在");
        }
        Settings settings = define.getSettings();
        if (Objects.isNull(settings)) {
            return null;
        }
        PrintConfig printConfig = settings.getPrint();
        Boolean enable = printConfig.getEnable();
        if (Objects.isNull(enable) || !enable) {
            // return null;
        }
        Map<String, Object> variables = this.getVariables(instanceId);
        User user = userService.getByUsername(instance.getStartUserId());
        variables.put("_initiator_", user.getName());
        variables.put("_code_", instance.getId());
        variables.put("_title_", instance.getName());
        String _startTime_ = Objects.nonNull(instance.getStartTime()) ?
                instance.getStartTime().format(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")) :
                "";
        variables.put("_startTime_", _startTime_);
        String _endTime_ = Objects.nonNull(instance.getEndTime()) ?
                instance.getEndTime().format(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")) :
                "";
        variables.put("_endTime_", _endTime_);
        String _currentTime_ = LocalDateTime.now().format(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss"));
        variables.put("_currentTime_", _currentTime_);
        variables.put("_status_", instance.getStatus().getDescription());
        SpreadSheet spreadSheet = printConfig.getTemplate();
        String json = objectMapper.writeValueAsString(spreadSheet);
        String format = StrUtil.format(json, variables);
        spreadSheet = objectMapper.readValue(format, SpreadSheet.class);
        return spreadSheet;
    }

    @Override
    public Boolean getPrintExist(String instanceId) {
        FlowInstance instance = flowInstanceDao.getInstance(instanceId);
        String definitionId = instance.getDefinitionId();
        FlowDefine define = flowDefineService.getDefine(definitionId);
        if (Objects.isNull(define)) {
            throw new BaseException("流程不存在");
        }
        Settings settings = define.getSettings();
        PrintConfig printConfig = settings.getPrint();
        return printConfig.getEnable();
    }
}
